# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
import pickle
from base64 import b64encode
from typing import Any
from typing import List
from typing import Set
from typing import Tuple

import numpy as np
import pandas as pd
from dacite import from_dict

from kolena._api.v1.generic import Search as API
from kolena._api.v2.search import Path as PATH_V2
from kolena._api.v2.search import UploadDatasetEmbeddingsRequest
from kolena._api.v2.search import UploadDatasetEmbeddingsResponse
from kolena._experimental.search._internal.datatypes import DatasetEmbeddingsDataFrameSchema
from kolena._experimental.search._internal.datatypes import LocatorEmbeddingsDataFrameSchema
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.state import API_V2
from kolena.dataset._common import COL_DATAPOINT_ID_OBJECT
from kolena.dataset._common import validate_dataframe_ids
from kolena.dataset.dataset import _load_dataset_metadata
from kolena.dataset.dataset import _to_serialized_dataframe
from kolena.errors import InputValidationError
from kolena.errors import NotFoundError


def upload_embeddings(key: str, embeddings: List[Tuple[str, np.ndarray]]) -> None:
    """
    Upload a list of search embeddings corresponding to sample locators.

    :param key: String value uniquely corresponding to the model used to extract the embedding vectors.
        This is typically a locator.
    :param embeddings: List of locator-embedding pairs, as tuples. Locators should be string values, while embeddings
        should be an `numpy.typing.ArrayLike` of numeric values.
    :raises InputValidationError: The provided embeddings input is not of a valid format
    """
    init_response = init_upload()
    locators, search_embeddings = [], []
    for locator, embedding in embeddings:
        if not np.issubdtype(embedding.dtype, np.number):
            raise InputValidationError("unexpected non-numeric embedding dtype")
        locators.append(locator)
        search_embeddings.append(b64encode(pickle.dumps(embedding.astype(np.float32))).decode("utf-8"))
    df_embeddings = pd.DataFrame(dict(key=[key] * len(embeddings), locator=locators, embedding=search_embeddings))
    df_validated = validate_df_schema(df_embeddings, LocatorEmbeddingsDataFrameSchema)

    log.info(f"uploading embeddings for key '{key}'")
    upload_data_frame(df=df_validated, load_uuid=init_response.uuid)
    request = API.UploadEmbeddingsRequest(
        uuid=init_response.uuid,
    )
    res = krequests.post(
        endpoint_path=API.Path.EMBEDDINGS.value,
        data=json.dumps(dataclasses.asdict(request)),
    )
    krequests.raise_for_status(res)
    data = from_dict(data_class=API.UploadEmbeddingsResponse, data=res.json())
    log.success(f"uploaded embeddings for key '{key}' on {data.n_samples} samples")


def upload_dataset_embeddings(dataset_name: str, key: str, df_embedding: pd.DataFrame) -> None:
    """
    Upload a list of search embeddings for a dataset.

    :param dataset_name: String value indicating the name of the dataset for which the embeddings will be uploaded.
    :param key: String value uniquely corresponding to the model used to extract the embedding vectors.
        This is typically a locator.
    :param df_embedding: Dataframe containing id fields for identifying datapoints in the dataset and the associated
        embeddings as `numpy.typing.ArrayLike` of numeric values.
    :raises NotFoundError: The given dataset does not exist.
    :raises InputValidationError: The provided input is not valid.
    """

    embedding_lengths: Set[int] = set()

    def encode_embedding(embedding: Any) -> str:
        if not np.issubdtype(embedding.dtype, np.number):
            raise InputValidationError("unexpected non-numeric embedding dtype")
        embedding_lengths.add(len(embedding))
        return b64encode(pickle.dumps(embedding.astype(np.float32))).decode("utf-8")

    # encode embeddings to string
    df_embedding["embedding"] = df_embedding["embedding"].apply(encode_embedding)
    if len(embedding_lengths) > 1:
        raise InputValidationError(f"embeddings are not of the same size, found {embedding_lengths}")

    # prepare the id objects of the dataset
    existing_dataset = _load_dataset_metadata(dataset_name)
    if not existing_dataset:
        raise NotFoundError(f"dataset {dataset_name} does not exist")
    id_fields = existing_dataset.id_fields
    validate_dataframe_ids(df_embedding, id_fields)
    df_serialized_datapoint_id_object = _to_serialized_dataframe(
        df_embedding[sorted(id_fields)],
        column=COL_DATAPOINT_ID_OBJECT,
    )
    df_embedding = pd.concat([df_embedding, df_serialized_datapoint_id_object], axis=1)

    df_embedding["key"] = key
    df_validated = validate_df_schema(df_embedding, DatasetEmbeddingsDataFrameSchema)

    log.info(f"uploading embeddings for dataset '{dataset_name}' and key '{key}'")
    init_response = init_upload()
    upload_data_frame(df=df_validated, load_uuid=init_response.uuid)
    request = UploadDatasetEmbeddingsRequest(
        uuid=init_response.uuid,
        name=dataset_name,
    )
    res = krequests.post(
        endpoint_path=PATH_V2.EMBEDDINGS.value,
        api_version=API_V2,
        data=json.dumps(dataclasses.asdict(request)),
    )
    krequests.raise_for_status(res)
    data = from_dict(data_class=UploadDatasetEmbeddingsResponse, data=res.json())
    log.success(f"uploaded embeddings for dataset '{dataset_name}' and key '{key}' on {data.n_datapoints} datapoints")



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noreorder
from kolena._utils.datatypes import DataObject
from .test_sample import Audio
from .test_sample import Metadata
from .test_sample import Image
from .test_sample import ImagePair
from .test_sample import ImageText
from .test_sample import TestSample
from .test_sample import Composite
from .test_sample import Text
from .test_sample import BaseVideo
from .test_sample import Video
from .test_sample import Document
from .test_sample import PointCloud
from .ground_truth import GroundTruth
from .inference import Inference
from .workflow import Workflow
from .test_case import TestCase
from .test_suite import TestSuite
from .model import Model
from .plot import AxisConfig
from .plot import Plot
from .plot import Curve
from .plot import CurvePlot
from .plot import ConfusionMatrix
from .plot import Histogram
from .plot import BarPlot
from .evaluator import MetricsTestCase
from .evaluator import MetricsTestSample
from .evaluator import MetricsTestSuite
from .evaluator import Evaluator
from .evaluator import EvaluatorConfiguration
from .evaluator_function import BasicEvaluatorFunction
from .evaluator_function import TestCases
from .evaluator_function import EvaluationResults
from .evaluator_function import no_op_evaluator
from .test_run import TestRun
from .test_run import test
from .define_workflow import define_workflow

__all__ = [
    "DataObject",
    "Metadata",
    "Image",
    "ImagePair",
    "ImageText",
    "TestSample",
    "Composite",
    "Text",
    "BaseVideo",
    "Video",
    "Document",
    "PointCloud",
    "Audio",
    "GroundTruth",
    "Inference",
    "Workflow",
    "TestCase",
    "TestSuite",
    "Model",
    "AxisConfig",
    "Plot",
    "Curve",
    "CurvePlot",
    "ConfusionMatrix",
    "Histogram",
    "BarPlot",
    "MetricsTestCase",
    "MetricsTestSample",
    "MetricsTestSuite",
    "Evaluator",
    "EvaluatorConfiguration",
    "BasicEvaluatorFunction",
    "TestCases",
    "EvaluationResults",
    "no_op_evaluator",
    "TestRun",
    "test",
    "define_workflow",
]



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import Counter
from typing import List
from typing import Optional

import pandas as pd

from kolena._utils import log
from kolena.errors import DuplicateDatapointIdError
from kolena.errors import InputValidationError

COL_DATAPOINT = "datapoint"
COL_DATAPOINT_ID_OBJECT = "datapoint_id_object"
COL_EVAL_CONFIG = "eval_config"
COL_RESULT = "result"
COL_THRESHOLDED_OBJECT = "thresholded_object"
_MAX_DUPLICATE_ID_REPORT = 10

DEFAULT_SOURCES = [dict(type="sdk")]


def validate_batch_size(batch_size: int) -> None:
    if batch_size <= 0:
        raise InputValidationError(f"invalid batch_size '{batch_size}': expected positive integer")


def validate_id_fields(id_fields: List[str], existing_id_fields: Optional[List[str]] = None) -> None:
    if len(id_fields) == 0:
        raise InputValidationError("invalid id_fields: expected at least one field")
    if len(Counter(id_fields)) != len(id_fields):
        raise InputValidationError(
            "invalid id_fields: id fields should not contain duplicates",
        )
    if existing_id_fields:
        if set(id_fields) != set(existing_id_fields):
            log.warn(
                f"ID field for the existing dataset has been changed from {existing_id_fields} to {id_fields},"
                f" this will disassociate the existing result from the new datapoints",
            )


def _validate_dataframe_ids_uniqueness(df: pd.DataFrame, id_fields: List[str]) -> None:
    try:
        duplicates = df[id_fields].duplicated()
        if duplicates.any():
            duplicate_ids = df[duplicates][id_fields]
            duplicate_ids.drop_duplicates(inplace=True)
            raise DuplicateDatapointIdError(
                f"invalid id_fields: input dataframe's id field values are not unique for {id_fields}",
                duplicate_ids=duplicate_ids[:_MAX_DUPLICATE_ID_REPORT].to_dict(orient="records"),
            )
    except TypeError as e:
        raise InputValidationError("id fields must be hashable") from e


def validate_dataframe_ids(df: pd.DataFrame, id_fields: List[str]) -> None:
    for id_field in id_fields:
        if id_field not in df.columns:
            raise InputValidationError(
                f"invalid id_fields: field '{id_field}' does not exist in dataframe",
            )
    _validate_dataframe_ids_uniqueness(df, id_fields)


def validate_dataframe_have_other_columns_besides_ids(df: pd.DataFrame, id_fields: List[str]) -> None:
    if set(df.columns) == set(id_fields):
        raise InputValidationError("dataframe only contains id fields")
